import argparse
import os
import sys
from pyspark.sql.types import *
from pyspark.sql import SparkSession
from pyspark.sql.window import Window
import pyspark.sql.functions as F

from lkb.factory import TRENDNet_Concept
from lkb.lkb_base_class import Conceptual_Edge, Conceptual_Relation, Lexical_Edge, Term
from lkb.basic_lkb import Basic_Lexical_Knowledge_Base
from .import_umls import download_umls


def getspark():
    return SparkSession.builder.master("local").appName("local tests").getOrCreate()


spark = getspark()

UMLS_META_PATH = "/Users/coltonflowers/Work/TREND_repos/krystie/lkb/2021AB/META"


def derive_trendnet(umls_meta_path, output_path):

    mrconso_schema = StructType(
        [
            StructField("CUI", StringType(), False),
            StructField("LAT", StringType(), False),
            StructField("TS", StringType(), False),
            StructField("LUI", StringType(), False),
            StructField("STT", StringType(), False),
            StructField("SUI", StringType(), False),
            StructField("ISPREF", StringType(), False),
            StructField("AUI", StringType(), False),
            StructField("SAUI", StringType(), True),
            StructField("SCUI", StringType(), True),
            StructField("SDUI", StringType(), True),
            StructField("SAB", StringType(), False),
            StructField("TTY", StringType(), False),
            StructField("CODE", StringType(), False),
            StructField("STR", StringType(), False),
            StructField("SRL", IntegerType(), False),
            StructField("SUPPRESS", StringType(), False),
            StructField("CVF", IntegerType(), False),
        ]
    )

    mrconso = (
        spark.read.options(
            delimiter="|",
            nullValues="",
        )
        .schema(mrconso_schema)
        .csv(f'file:{os.path.join(umls_meta_path, "MRCONSO.RRF")}')
    )

    mrrel_schema = StructType(
        [
            StructField("CUI1", StringType(), False),
            StructField("AUI1", StringType(), True),
            StructField("STYPE1", StringType(), False),
            StructField("REL", StringType(), False),
            StructField("CUI2", StringType(), False),
            StructField("AUI2", StringType(), True),
            StructField("STYPE2", StringType(), False),
            StructField("RELA", StringType(), True),
            StructField("RUI", StringType(), False),
            StructField("SRUI", StringType(), True),
            StructField("SAB", StringType(), False),
            StructField("SL", StringType(), False),
            StructField("RG", StringType(), True),
            StructField("DIR", StringType(), True),
            StructField("SUPPRESS", StringType(), False),
            StructField("CVF", IntegerType(), True),
        ]
    )
    mrrel = (
        spark.read.options(
            delimiter="|",
            nullValues="",
        )
        .schema(mrrel_schema)
        .csv(f'file:{os.path.join(umls_meta_path, "MRREL.RRF")}')
    )
    mrsat_schema = StructType(
        [
            StructField("CUI", StringType(), False),
            StructField("LUI", StringType(), True),
            StructField("SUI", StringType(), True),
            StructField("METAUI", StringType(), True),
            StructField("STYPE", StringType(), False),
            StructField("CODE", StringType(), True),
            StructField("ATUI", StringType(), False),
            StructField("SATUI", StringType(), True),
            StructField("ATN", StringType(), False),
            StructField("SAB", StringType(), False),
            StructField("ATV", StringType(), True),
            StructField("SUPPRESS", StringType(), False),
            StructField("CVF", IntegerType(), True),
        ]
    )

    mrsat = (
        spark.read.options(
            delimiter="|",
            nullValues="",
        )
        .schema(mrsat_schema)
        .csv(f'file:{os.path.join(umls_meta_path, "MRSAT.RRF")}')
    )

    mrdef_schema = StructType(
        [
            StructField("CUI", StringType(), False),
            StructField("AUI", StringType(), False),
            StructField("ATUI", StringType(), False),
            StructField("SATUI", StringType(), True),
            StructField("SAB", StringType(), False),
            StructField("DEF", StringType(), False),
            StructField("SUPPRESS", StringType(), False),
            StructField("CVF", IntegerType(), True),
        ]
    )
    mrdef = (
        spark.read.options(delimiter="|", nullValues="")
        .schema(mrdef_schema)
        .csv(f'file:{os.path.join(umls_meta_path, "MRDEF.RRF")}')
    )

    mrsty_schema = StructType(
        [
            StructField("CUI", StringType(), False),
            StructField("TUI", StringType(), False),
            StructField("STN", StringType(), False),
            StructField("STY", StringType(), False),
            StructField("ATUI", StringType(), False),
            StructField("CVF", IntegerType(), True),
        ]
    )

    mrsty = (
        spark.read.options(
            delimiter="|",
            nullValues="",
        )
        .schema(mrsty_schema)
        .csv(f'file:{os.path.join(umls_meta_path, "MRDEF.RRF")}')
    )
    mrconso.createOrReplaceTempView("mrconso")
    mrsat.createOrReplaceTempView("mrsat")

    canonicalCond = (
        (F.col("STT") == "PF")
        & (F.col("TS") == "P")
        & (F.col("ISPREF") == "Y")
        & (F.col("LAT") == "ENG")
    )

    concept_names = (
        mrconso.select("CUI", "STR")
        .orderBy(canonicalCond, F.col("STR"))
        .groupBy("CUI")
        .agg(F.last("STR").alias("CONCEPT_NAME"))
    )

    definitions = (
        mrdef.where(
            ((F.col("SAB") == "RXNORM") | (F.col("SAB") == "SNOMEDCT_US"))
            | (F.col("SAB") == "MSH")
        )
        .select(F.col("CUI"), F.col("DEF").alias("definition"))
        .dropDuplicates()
    )
    semantic_types = mrsty.select(F.col("CUI"), F.col("STY").alias("semantic_type"))

    linkable_brandname_concepts = (
        mrrel.where(F.col("RELA") == "has_tradename")
        .join(
            mrconso.select(F.col("CUI").alias("CUI2")).where(
                F.col("SAB") == "SNOMEDCT_US"
            )
        )
        .select(F.col("CUI1").alias("brandname_CUI"))
        .dropDuplicates()
    )

    unhealed_trendnet_atoms = (
        mrconso.join(
            linkable_brandname_concepts,
            linkable_brandname_concepts.brandname_CUI == mrconso.CUI,
            "left_outer",
        )
        .withColumn(
            "RXNORM_BRANDNAME_CONCEPT",
            (
                F.array_contains(
                    F.collect_set("SAB").over(Window.partitionBy("CUI")), "RXNORM"
                )
                & F.array_contains(
                    F.collect_set("TTY").over(Window.partitionBy("CUI")), "BN"
                )
            ),
        )
        .withColumn(
            "LINKABLE_BRANDNAME",
            (F.col("brandname_CUI").isNotNull() & F.col("RXNORM_BRANDNAME_CONCEPT")),
        )
        .drop("brandname_CUI")
        .withColumn("SNOMED_ATOM", F.col("SAB") == "SNOMEDCT_US")
        .withColumn(
            "CONCEPT_CONTAINS_SNOMED_ATOM",
            F.array_contains(
                F.collect_set("SAB").over(Window.partitionBy("CUI")), "SNOMEDCT_US"
            ),
        )
        .drop("SNOMED_ATOM")
        .withColumn(
            "CONCEPT_CONTAINS_CHV_ATOM",
            F.array_contains(
                F.collect_set("SAB").over(Window.partitionBy("CUI")), "CHV"
            ),
        )
        .withColumn(
            "UNFILTERED_TRENDNET_CONCEPT",
            (F.col("CONCEPT_CONTAINS_SNOMED_ATOM") | F.col("LINKABLE_BRANDNAME")),
        )
        .withColumn(
            "TRENDNET_CONCEPT",
            (F.col("CONCEPT_CONTAINS_SNOMED_ATOM") | F.col("LINKABLE_BRANDNAME"))
            & F.col("CONCEPT_CONTAINS_CHV_ATOM"),
        )
        .join(concept_names, "CUI")
        .where(F.col("UNFILTERED_TRENDNET_CONCEPT"))
        .join(definitions, "CUI", "left")
        .join(semantic_types, "CUI", "left")
    ).cache()

    # HEAL THE HIERARCHY
    in_chv = {
        row["CUI"]: row["CONCEPT_CONTAINS_CHV_ATOM"]
        for row in unhealed_trendnet_atoms.select(
            "CUI", "CONCEPT_CONTAINS_CHV_ATOM"
        ).collect()
    }

    hierarchy_edges = [
        (row["CUI1"], row["CUI2"])
        for row in (
            mrrel.where(
                (F.col("SAB") == "SNOMEDCT_US") & (F.col("RELA") == "inverse_isa")
            )
            .join(
                unhealed_trendnet_atoms.select(F.col("CUI").alias("CUI1")),
                "CUI1",
                "leftsemi",
            )
            .join(
                unhealed_trendnet_atoms.select(F.col("CUI").alias("CUI2")),
                "CUI2",
                "leftsemi",
            )
        ).collect()
    ]

    graph = {}
    for edge in hierarchy_edges:
        source_cui = edge[1]
        dest_cui = edge[0]
        if source_cui not in graph:
            graph[source_cui] = []
        graph[source_cui].append(dest_cui)
    print(len(hierarchy_edges))
    number_of_edges_traversed = 0
    top_concept = "C2720507"  # SNOMED CT Concept (SNOMED RT+CTV3)
    concepts = [top_concept]
    healed_edges = []
    last_alive_parent = {"C2720507": "TOP_CONCEPT"}
    traversed = set()
    while concepts:
        hierarchy_edges
        parent_concept = concepts.pop()
        if parent_concept in graph and parent_concept not in traversed:
            traversed.add(parent_concept)
            children_concepts = graph[parent_concept]
            number_of_edges_traversed += len(children_concepts)
            if number_of_edges_traversed % 1000 == 0:
                print(f"number_of_edges_traversed:{number_of_edges_traversed}")
            for child_concept in children_concepts:
                if not in_chv[parent_concept]:
                    if not in_chv[child_concept]:
                        last_alive_parent[child_concept] = last_alive_parent[
                            parent_concept
                        ]
                    elif last_alive_parent[parent_concept] != "TOP_CONCEPT":
                        healed_edges.append(
                            (last_alive_parent[parent_concept], child_concept)
                        )
                else:
                    if not in_chv[child_concept]:
                        last_alive_parent[child_concept] = parent_concept
                    elif parent_concept != "TOP_CONCEPT":
                        healed_edges.append((parent_concept, child_concept))
            concepts.extend(children_concepts)

    healed_edges = set(healed_edges)

    trendnet_atoms = unhealed_trendnet_atoms.where(F.col("TRENDNET_CONCEPT"))
    nonhierarchical_trendnet_relations = (
        mrrel.where(
            (F.col("RELA").isin(["has_tradename", "tradename_of"]))
            | (
                (F.col("SAB") == "SNOMEDCT_US")
                & (
                    ~F.col("RELA").isin(
                        [
                            "mapped_to",
                            "isa",
                            "inverse_isa",
                            "mapped_from",
                            "inverse_was_a",
                            "same_as",
                            "replaced_by",
                            "was_a",
                            "possibly_equivalent_to",
                        ]
                    )
                    & (F.col("CUI1") != F.col("CUI2"))
                )
            )
        )
        .join(trendnet_atoms.withColumnRenamed("CUI", "CUI1"), "CUI1", "leftsemi")
        .join(trendnet_atoms.withColumnRenamed("CUI", "CUI2"), "CUI2", "leftsemi")
    )

    healed_edges = set(healed_edges)

    ####################################################################################################################################################

    ##########################################GET CONCEPTS+RELATIONS####################################################################################

    ####################################################################################################################################################

    concepts = trendnet_atoms.select(
        F.col("CUI").alias("id"),
        F.col("CONCEPT_NAME").alias("concept_name"),
        F.col("definition"),
        F.col("semantic_type"),
    ).dropDuplicates()
    nonhierarchical_conceptual_relations = (
        nonhierarchical_trendnet_relations.select(F.col("RELA").alias("string"))
        .dropDuplicates()
        .withColumn("id", F.concat(F.lit("R"), F.monotonically_increasing_id()))
    )
    nonhierarchical_conceptual_edges = (
        nonhierarchical_trendnet_relations.join(
            nonhierarchical_conceptual_relations.withColumnRenamed("string", "rela"),
            "rela",
        ).select(
            F.col("id").alias("rel_id"),
            F.col("CUI2").alias("concept_id_1"),
            F.col("CUI1").alias("concept_id_2"),
        )
    ).dropDuplicates()

    ####################################################################################################################################################

    ##########################################GET TERMS#################################################################################################

    ####################################################################################################################################################

    chv_trendnet_atoms = trendnet_atoms.where(F.col("SAB") == "CHV")
    preferred_terms = chv_trendnet_atoms.where(F.col("TTY") == "PT").select(
        "CUI",
        F.array_sort(F.split(F.col("STR"), " ")).alias("preferred_term_normalized"),
    )
    _lexical_edges = (
        chv_trendnet_atoms.join(preferred_terms, "CUI")
        .where(
            (
                F.array_sort(F.split(F.col("STR"), " "))
                != F.col("preferred_term_normalized")
            )
            | (F.col("TTY") == "PT")
        )
        .select(
            F.col("CUI").alias("concept_id"),
            F.col("SUI").alias("term_id"),
            F.col("STR").alias("string"),
        )
        .dropDuplicates()
    )

    lexical_edges = _lexical_edges.select("concept_id", "term_id")

    terms = _lexical_edges.select(
        F.col("term_id").alias("id"), "string"
    ).dropDuplicates()

    #   display(chv_trendnet_atoms.where(F.col("CUI")=="C1170371"))
    concepts = [
        TRENDNet_Concept(**{key: value for key, value in row.asDict().items() if value})
        for row in concepts.collect()
    ]
    lexical_edges = [Lexical_Edge(**row.asDict()) for row in lexical_edges.collect()]
    terms = [Term(**row.asDict()) for row in terms.collect()]
    conceptual_relations = [
        Conceptual_Relation(**row.asDict())
        for row in nonhierarchical_conceptual_relations.collect()
    ] + [Conceptual_Relation("-1", "isa"), Conceptual_Relation("-2", "inverse_isa")]
    conceptual_edges = (
        [
            Conceptual_Edge(**row.asDict())
            for row in nonhierarchical_conceptual_edges.collect()
        ]
        + [
            Conceptual_Edge(rel_id="-1", concept_id_1=edge[1], concept_id_2=edge[0])
            for edge in healed_edges
            if edge[0] != edge[1]
        ]
        + [
            Conceptual_Edge(rel_id="-2", concept_id_1=edge[0], concept_id_2=edge[1])
            for edge in healed_edges
            if edge[0] != edge[1]
        ]
    )

    Basic_Lexical_Knowledge_Base(
        concepts=concepts,
        terms=terms,
        conceptual_edges=conceptual_edges,
        lexical_edges=lexical_edges,
        conceptual_relations=conceptual_relations,
    ).write_json(output_path)


def main():
    parser = argparse.ArgumentParser(description="process user given parameters")
    parser.add_argument(
        "-u",
        "--umlspath",
        required=False,
        dest="umlspath",
        help="enter path to umls zip file",
    )
    parser.add_argument(
        "-d",
        "--databricks",
        action="store_true",
        dest="databricks",
        help="indicate if we are in a databricks cluster",
    )
    parser.add_argument(
        "-k",
        "--apikey",
        required=False,
        dest="apikey",
        help="enter api key from your UTS Profile",
    )
    parser.add_argument(
        "-v",
        "--version",
        required=False,
        dest="version",
        default="current",
        help="enter version example-2015AA",
    )
    parser.add_argument(
        "-f",
        "--output_path",
        required=True,
        dest="output_path",
        help="enter a name for your output path for trendnet",
    )
    args = parser.parse_args()
    if not args.umlspath:
        if not args.databricks:
            if not args.apikey:
                sys.exit("Please enter your api key.")
            else:
                apikey = args.apikey
        else:
            import dbutils

            apikey = dbutils.secrets.get(scope="umls", key="api_key")
        download_umls(apikey, ".")
        umls_metathesaurus_path = os.path.join(".", "2021/META")
    else:
        umls_metathesaurus_path = args.umlspath
    if not args.output_path:
        sys.exit("Please enter an output directory.")
    derive_trendnet(umls_metathesaurus_path, args.output_path)
